#include <stdio.h>
#include <memory.h>

#include <Application.h>
#include <OS.h>
#include <String.h>
#include <Screen.h>
#include <File.h>
#include <TranslatorRoster.h>
#include <BitmapStream.h>
#include <Entry.h>

#include <new>
#include <iostream>
#include <vector>

#include "app_params.h"
#include "setup.h"

#include "frmNull.h"
#include "frmBitmap.h"
#include "frmDirectWindow.h"
#include "frmDirectWindowLC.h"
#include "frmWindowScreen.h"

#include "rend.h"

#define	DEFAULT_NUM_RING_BUFFERS	2
#define DEFAULT_FRAMEWORK_TYPE		FRAMEWORK_BVIEW
#define DEFAULT_SAMPLING			RENDER_STATE_SAMP_CONST_NEAREST

#define CONTENT_AREA_DIM_X			512
#define CONTENT_AREA_DIM_Y			384

// application sub-class(es) /////////////////////////////////////////////////////

class App : BApplication
{
	status_t				mAppStatus;
	BString					mErrorString;

	Framework *				mFramework;
	AppParams				mParams;

	bool					mSkipSetup;
	bool					mDumpStats;

	static int32			DrawingThread(void *data);
	thread_id				mDrawThreadID; 

	bigtime_t				mElapsedTime;
	uint32					mFrameCount;

	std::vector<BBitmap *>	mTextureSet[2];
	volatile int32			mActiveTexSet;

	void					WipeTexSet(int32 set)
							{ while (mTextureSet[set].size()) { delete mTextureSet[set].back(); mTextureSet[set].pop_back(); } }
public:
	App(status_t&);
	~App();

	virtual void			ArgvReceived(int32 argc, char **argv);		
	virtual void			RefsReceived(BMessage *message);
	virtual void			ReadyToRun();
	virtual bool			QuitRequested();

private:
	void					CreateDrawingThread_BWindow();
};

App::App(status_t& error) :
	BApplication("application/x-blu-framedemo", &error)
{
	if (B_OK != (mAppStatus = error))
		mErrorString = "failure at BApplication construction.";

	mSkipSetup = false;
	mDumpStats = false;

	mParams.frameworkType					= DEFAULT_FRAMEWORK_TYPE;
	mParams.frameworkParams.numRingBuffers	= DEFAULT_NUM_RING_BUFFERS;
	mParams.frameworkParams.pixelFormat		= B_NO_COLOR_SPACE;
	mParams.frameworkParams.frameBounds		= BPoint(CONTENT_AREA_DIM_X - 1, CONTENT_AREA_DIM_Y - 1);

	mParams.rendSampling					= DEFAULT_SAMPLING;

	mFramework = NULL;
	mFrameCount = 0;
	
	mActiveTexSet = 0;
}

App::~App()
{
	WipeTexSet(mActiveTexSet);

	if (B_OK == mAppStatus)
	{
		if (mDumpStats)
		{
			uint32 i = 0;
			while ((i < AppParams::nPixFormats) && (AppParams::pixFmtId[i].space != mParams.frameworkParams.pixelFormat))
				i++;

			std::cout << "number of ring buffers " << mParams.frameworkParams.numRingBuffers << std::endl
					  << "pixel format " << AppParams::pixFmtId[i].textual << std::endl
					  << "elapsed time (sec) " << ((double)mElapsedTime / 1000000.0) << std::endl
					  << "drawn frames " << mFrameCount << std::endl
					  << "framerate (fps) " << ((mFrameCount / (double)mElapsedTime) * 1000000.0) << std::endl;
		}
	}
	else
		std::cerr << mErrorString.String() << std::endl;
}

void App::ArgvReceived(int32 argc, char **argv)
{
	BMessage messageRefs(B_REFS_RECEIVED);

	for (int32 i = 1; (B_OK == mAppStatus) && (i < argc); i++)
	{
		if (! ::strcmp(argv[i], PARA_STATS))
		{
			mDumpStats = true;
			continue;
		}

		if (! ::strcmp(argv[i], PARA_NOSETUP))
		{
			mSkipSetup = true;
			continue;
		}

		if (! ::strcmp(argv[i], PARA_FRAMEWORK))
		{
			if (argc > ++i)
			{
				if (! ::strcmp(argv[i], PARA_FRAMEWORK_NULL))
				{
					mParams.frameworkType = FRAMEWORK_NULL;
					continue;
				}
				if (! ::strcmp(argv[i], PARA_FRAMEWORK_BVIEW))
				{
					mParams.frameworkType = FRAMEWORK_BVIEW;
					continue;
				}
				if (! ::strcmp(argv[i], PARA_FRAMEWORK_BDWIN))
				{
					mParams.frameworkType = FRAMEWORK_BDIRECTWINDOW;
					continue;
				}
				if (! ::strcmp(argv[i], PARA_FRAMEWORK_BDWINLC))
				{
					mParams.frameworkType = FRAMEWORK_BDIRECTWINDOW_LC;
					continue;
				}
				if (! ::strcmp(argv[i], PARA_FRAMEWORK_BWINSCR))
				{
					mParams.frameworkType = FRAMEWORK_BWINDOWSCREEN;
					continue;
				}
			}
		}

		if (! ::strcmp(argv[i], PARA_NUMBUFFS))
		{
			if (argc > ++i)
			{
				if (1 == ::sscanf(argv[i], "%ld", &mParams.frameworkParams.numRingBuffers) &&
					MIN_NUM_RING_BUFFERS <= mParams.frameworkParams.numRingBuffers &&
					MAX_NUM_RING_BUFFERS >= mParams.frameworkParams.numRingBuffers)
					continue;
			}
		}

		if (! ::strcmp(argv[i], PARA_TEXTUREFILE)) 
		{
			if (argc > ++i)
			{
				entry_ref ref;

				BEntry entry(argv[i]);
				entry.GetRef(&ref);

				messageRefs.AddRef("refs", &ref);
				continue;
			}
		}

		if (! ::strcmp(argv[i], PARA_SAMPLING))
		{
			if (argc > ++i)
			{
				if (! ::strcmp(argv[i], PARA_SAMPLING_CONST_NEAREST))
				{
					mParams.rendSampling = RENDER_STATE_SAMP_CONST_NEAREST;
					continue;
				}
				if (! ::strcmp(argv[i], PARA_SAMPLING_CONST_LINEAR))
				{
					mParams.rendSampling = RENDER_STATE_SAMP_CONST_LINEAR;
					continue;
				}
				if (! ::strcmp(argv[i], PARA_SAMPLING_NEAREST_NEAREST))
				{
					mParams.rendSampling = RENDER_STATE_SAMP_NEAREST_NEAREST;
					continue;
				}
				if (! ::strcmp(argv[i], PARA_SAMPLING_NEAREST_LINEAR))
				{
					mParams.rendSampling = RENDER_STATE_SAMP_NEAREST_LINEAR;
					continue;
				}
				if (! ::strcmp(argv[i], PARA_SAMPLING_LINEAR_NEAREST))
				{
					mParams.rendSampling = RENDER_STATE_SAMP_LINEAR_NEAREST;
					continue;
				}
				if (! ::strcmp(argv[i], PARA_SAMPLING_LINEAR_LINEAR))
				{
					mParams.rendSampling = RENDER_STATE_SAMP_LINEAR_LINEAR;
					continue;
				}
			}
		}

		// this argument is invalid, hint the user of app's command-line usage

		mErrorString << "usage: "
			<< (const char *) argv[0] << " [OPTION]..\n"
			<< "OPTION\n"
			<< "\t" << PARA_NOSETUP << "\t\tbypass the setup at startup\n"
			<< "\t" << PARA_NUMBUFFS << " NUM\t\tnumber of ring buffers, domain of [" << (int32) MIN_NUM_RING_BUFFERS
			<< ", " << (int32) MAX_NUM_RING_BUFFERS << "], default is " << DEFAULT_NUM_RING_BUFFERS << "\n"
			<< "\t" << PARA_FRAMEWORK << " FWTYPE\tuse the specified framebuffer framework\n"
			<< "\t" << PARA_TEXTUREFILE << " FILE\t\tuse texture from specified file\n"
			<< "\t" << PARA_SAMPLING << " SAMPTYPE\tuse specified texture sampling\n"
			<< "\t" << PARA_STATS << "\t\t\toutput runtime statistics at successfull completion\n"
			<< "FWTYPE\n"
			<< "\t" << PARA_FRAMEWORK_NULL << "\t\t\t" << DESC_FRAMEWORK_NULL << "\n"
			<< "\t" << PARA_FRAMEWORK_BVIEW << "\t\t\t" << DESC_FRAMEWORK_BVIEW << "\n"
			<< "\t" << PARA_FRAMEWORK_BDWIN << "\t\t" << DESC_FRAMEWORK_BDWIN << "\n"
			<< "\t" << PARA_FRAMEWORK_BDWINLC << "\t\t" << DESC_FRAMEWORK_BDWINLC << "\n"
			<< "\t" << PARA_FRAMEWORK_BWINSCR << "\t\t" << DESC_FRAMEWORK_BWINSCR << "\n"
			<< "SAMPTYPE\n"
			<< "\t" << PARA_SAMPLING_CONST_NEAREST << "\t\t" << DESC_SAMPLING_CONST_NEAREST << "\n"
			<< "\t" << PARA_SAMPLING_CONST_LINEAR << "\t\t" << DESC_SAMPLING_CONST_LINEAR << "\n"
			<< "\t" << PARA_SAMPLING_NEAREST_NEAREST << "\t\t" << DESC_SAMPLING_NEAREST_NEAREST << "\n"
			<< "\t" << PARA_SAMPLING_NEAREST_LINEAR << "\t\t" << DESC_SAMPLING_NEAREST_LINEAR << "\n"
			<< "\t" << PARA_SAMPLING_LINEAR_NEAREST << "\t\t" << DESC_SAMPLING_LINEAR_NEAREST << "\n"
			<< "\t" << PARA_SAMPLING_LINEAR_LINEAR << "\t\t" << DESC_SAMPLING_LINEAR_LINEAR << "\n";

		mAppStatus = B_ERROR;
	}
	
	if (B_OK == mAppStatus)
		if (!messageRefs.IsEmpty())
			::be_app_messenger.SendMessage(&messageRefs);
}

void App::ReadyToRun(void)
{
	if (B_OK == mAppStatus)
	{
		if (mSkipSetup)	// setup interface supressed; set framework's pixel format to screen's current
		{
			BScreen *screen = new(std::nothrow) BScreen();
			
			if (screen != NULL && (*screen).IsValid())
			{
				mParams.frameworkParams.pixelFormat = (*screen).ColorSpace();
			}
			else
			{
				mErrorString = "failed to construct a BScreen; cannot get screen's current color space.";
				mAppStatus = B_ERROR;
			}

			delete screen;
		}
		else // invoke the setup interface 
		{
			SetupWindow *wndSetup = new(std::nothrow) SetupWindow(mParams);

			if (NULL != wndSetup)
			{
				Lifespan::lifesign_t sign = wndSetup->CreateLifesign();

				if (B_ERROR < sign)
				{
					wndSetup->Show();
					mAppStatus = Lifespan::WaitForDeath(sign);
				}
				else
				{
					wndSetup->LockLooper();
					wndSetup->Quit();

					mErrorString = "failed to obtain a lifesign for the setup window.";
					mAppStatus = (status_t) sign;
				}
			}
			else
			{
				mErrorString = "'bad_alloc' at setup window construction.";
				mAppStatus = B_ERROR;
			}
		}
	}

	if (B_OK == mAppStatus) // create the display framework as requested
	{
		switch (mParams.frameworkType)
		{
		case FRAMEWORK_NULL:

			mFramework = new(std::nothrow) FrmNullWindow(mParams.frameworkParams, mAppStatus, mErrorString);

			CreateDrawingThread_BWindow();
			break;

		case FRAMEWORK_BVIEW:

			mFramework = new(std::nothrow) FrmBmpWindow(mParams.frameworkParams, mAppStatus, mErrorString);

			CreateDrawingThread_BWindow();
			break;

		case FRAMEWORK_BDIRECTWINDOW:

			mFramework = new(std::nothrow) FrmDirectWindow(mParams.frameworkParams, mAppStatus, mErrorString);

			CreateDrawingThread_BWindow();
			break;

		case FRAMEWORK_BDIRECTWINDOW_LC:

			mFramework = new(std::nothrow) FrmDirectWindowLC(mParams.frameworkParams, mAppStatus, mErrorString);

			CreateDrawingThread_BWindow();
			break;

		case FRAMEWORK_BWINDOWSCREEN:

			mFramework = new(std::nothrow) FrmWindowScreen(mParams.frameworkParams, mAppStatus, mErrorString);

			CreateDrawingThread_BWindow();
			break;
		}
	}

	if (B_OK == mAppStatus && mTextureSet[mActiveTexSet].size() == 0)
	{
		const int32 deftex_dimx = 16;
		const int32 deftex_dimy = 16;

		static const uint8 deftex[deftex_dimy * deftex_dimx] =
		{
			0,	 0,   0,   0,	0,	 0,	  0,   0,	255, 255, 255, 255, 255, 255, 255, 255,
			0,	 0,	  0,   0,	0,	 0,	  0,   0,	255, 255, 255, 255, 255, 255, 255, 255,
			0,	 0,	  0,   0,	0,	 0,	  0,   0,	255, 255, 255, 255, 255, 255, 255, 255,
			0,	 0,	  0,   0,	0,	 0,	  0,   0,	255, 255, 255, 255, 255, 255, 255, 255,
			0,	 0,	  0,   0,	0,	 0,	  0,   0,	255, 255, 255, 255, 255, 255, 255, 255,
			0,	 0,	  0,   0,	0,	 0,	  0,   0,	255, 255, 255, 255, 255, 255, 255, 255,
			0,	 0,	  0,   0,	0,	 0,	  0,   0,	255, 255, 255, 255, 255, 255, 255, 255,
			0,	 0,	  0,   0,	0,	 0,	  0,   0,	255, 255, 255, 255, 255, 255, 255, 255,
			255, 255, 255, 255,	255, 255, 255, 255,	0,	 0,	  0,   0,	0,	 0,	  0,   0,
			255, 255, 255, 255,	255, 255, 255, 255,	0,	 0,	  0,   0,	0,	 0,	  0,   0,
			255, 255, 255, 255,	255, 255, 255, 255,	0,	 0,	  0,   0,	0,	 0,	  0,   0,
			255, 255, 255, 255,	255, 255, 255, 255,	0,	 0,	  0,   0,	0,	 0,	  0,   0,
			255, 255, 255, 255,	255, 255, 255, 255,	0,	 0,	  0,   0,	0,	 0,	  0,   0,
			255, 255, 255, 255,	255, 255, 255, 255,	0,	 0,	  0,   0,	0,	 0,	  0,   0,
			255, 255, 255, 255,	255, 255, 255, 255,	0,	 0,	  0,   0,	0,	 0,	  0,   0,
			255, 255, 255, 255,	255, 255, 255, 255,	0,	 0,	  0,   0,	0,	 0,	  0,   0
		};

		BBitmap *bitmap = new BBitmap(BRect(0, 0, deftex_dimx - 1, deftex_dimy - 1), B_RGB32);

		for (int32 y = 0; y < deftex_dimy; y++)
			bitmap->SetBits(deftex + deftex_dimx * y, deftex_dimx, bitmap->BytesPerRow() * y, B_CMAP8);

		mTextureSet[mActiveTexSet].push_back(bitmap);
	}

	if (B_OK == mAppStatus)
		::resume_thread(mDrawThreadID);
	else
		::be_app_messenger.SendMessage(B_QUIT_REQUESTED);
}

void App::RefsReceived(BMessage *message)
{
	if (B_OK != mAppStatus)
		return;

	int32 active_set = mActiveTexSet ^ 1;

	WipeTexSet(active_set);

	BTranslatorRoster *roster = BTranslatorRoster::Default(); 
	entry_ref ref;
	int32 i = 0;

	while(message->FindRef("refs", i++, &ref) == B_OK)
	{
		BEntry entry(&ref, true);

		if (! entry.Exists())
			continue;

		BFile file(&entry, B_READ_ONLY);
		BBitmapStream stream;

		if (roster->Translate(&file, NULL, NULL, &stream, B_TRANSLATOR_BITMAP) == B_OK)
		{
			BBitmap *bitmap_src = NULL;
			stream.DetachBitmap(&bitmap_src);

			color_space space_src = bitmap_src->ColorSpace();

			if (B_RGBA32 == space_src || B_RGB32 == space_src)
			{
				mTextureSet[active_set].push_back(bitmap_src);
			}
			else
			{
				BRect bounds(bitmap_src->Bounds());

				BBitmap *bitmap_dst = new BBitmap(bounds, B_RGBA32, true);
				BView *view = new BView(bounds, NULL, B_FOLLOW_NONE, 0);

				bitmap_dst->AddChild(view);
				view->LockLooper();
				view->DrawBitmap(bitmap_src, bounds);
				view->UnlockLooper();
				bitmap_dst->RemoveChild(view);

				delete view;
				delete bitmap_src;

				mTextureSet[active_set].push_back(bitmap_dst);
			}
		}
	}

	if (mTextureSet[active_set].size())
		mActiveTexSet = active_set;

	// note of importance:
	// we'd be in trouble if this handler got invoked more than once within the lifespan of a single frame.
	// but that is not supposed to happen.
}

bool App::QuitRequested(void)
{
	return BApplication::QuitRequested();
}

void App::CreateDrawingThread_BWindow()
{
	// try to down-cast the framework to a BWindow
	BWindow *window = dynamic_cast<BWindow *>(mFramework);

	if (NULL == window)
	{
		if (B_OK == mAppStatus)
		{
			mErrorString = "expected a BWindow-ascending framework, got something else.";
			mAppStatus = B_ERROR;
		}
		return;
	}

	if (B_OK == mAppStatus)
	{
		mDrawThreadID = ::spawn_thread(DrawingThread, Framework::__painter_name(), B_NORMAL_PRIORITY, (void *) this);

		if (B_ERROR < mDrawThreadID)
		{
			window->Show();
			return;
		}

		mErrorString = "failed to spawn a drawing thread.";
		mAppStatus = (status_t) mDrawThreadID;
	}

	// either a spawn failure or an earlier failure: destroy the framework
	window->LockLooper();
	window->Quit();
}

// rendition //////////////////////////////////////////////////////

static struct ColorFormatsMapper
{
	struct native2render_t
	{
		color_space			native;
		uint32				render;
	};

	static native2render_t	native2render[6];

	ColorFormatsMapper()
	{
		bool sorted = false;
		while(!sorted)
		{
			sorted = true;
			for (uint32 i = 0; i < sizeof(native2render) / sizeof(native2render_t) - 1; i++)
				if ((uint32)native2render[i].native > (uint32)native2render[i + 1].native)
				{
					native2render_t tmp = native2render[i];
					native2render[i] = native2render[i + 1];
					native2render[i + 1] = tmp;
					sorted = false;
				}
		}
	}

	uint32 NativeToRender(color_space searched)
	{
		int32 lower_idx = -1;
		int32 upper_idx = sizeof(native2render) / sizeof(native2render_t);

		int32 check_idx = (lower_idx + upper_idx) / 2;

		while (lower_idx < check_idx && check_idx < upper_idx)
			if ((uint32)native2render[check_idx].native < (uint32)searched)
			{
				lower_idx = check_idx;
				check_idx = (lower_idx + upper_idx) / 2;
			}
			else
			if ((uint32)native2render[check_idx].native > (uint32)searched)
			{
				upper_idx = check_idx;
				check_idx = (lower_idx + upper_idx) / 2;
			}
			else
				break;

		if (native2render[check_idx].native == searched)
			return native2render[check_idx].render;
		else
			return 0;
	}
}
cfmapper;

ColorFormatsMapper::native2render_t ColorFormatsMapper::native2render[] =
{
	{ B_RGB15,	RENDER_STATE_PIX_ARGB1555 },
	{ B_RGBA15,	RENDER_STATE_PIX_ARGB1555 },
	{ B_RGB32,	RENDER_STATE_PIX_ARGB8888 },
	{ B_RGBA32, RENDER_STATE_PIX_ARGB8888 },
	{ B_RGB16,	RENDER_STATE_PIX_RGB565 },
	{ B_RGB24,	RENDER_STATE_PIX_RGB888 }
};

int32 App::DrawingThread(void *data)
{
	App *app			= (App *) data;
	Framework *frmwrk	= (*app).mFramework;

	Framework::BUFFER_DESC desc;

	///////////////////////////////////////////////////////////////

	const Vector6 vt0 = { -1,  1,  1,   1,	0,	0 };
	const Vector6 vt1 = {  1,  1,  1,   1,	0,	0 };
	const Vector6 vt2 = { -1, -1,  1,   1,	0,	0 };
	const Vector6 vt3 = {  1, -1,  1,   1,	0,	0 };

	const Vector8 vc0 = { -1,  1,  1,   1,	0,	0,	0,	1 };
	const Vector8 vc1 = {  1,  1,  1,   1,	0,	1,	0,	0 };
	const Vector8 vc2 = { -1, -1,  1,   1,	0,	0,	1,	0 };
	const Vector8 vc3 = {  1, -1,  1,   1,	0,	0,	0,	1 };

	const float * squareA[2][3] =
	{
		{ vt0, vt1, vt2 },
		{ vt2, vt1, vt3 }
	};

	const float * squareB[2][3] =
	{
		{ vc0, vc1, vc2 },
		{ vc2, vc1, vc3 }
	};

	Matrix4	cube[6];

	// front
	SetMat4(cube[0],
			1, 0, 0, 0,
			0, 1, 0, 0,
			0, 0, 1, 0,
			0, 0, 0, 1);

	// back
	Rotate(cube[1], M_PI, 0, 1, 0);

	// left
	Rotate(cube[2], -M_PI_2, 0, 1, 0);

	// right
	Rotate(cube[3], M_PI_2, 0, 1, 0);

	// top
	Rotate(cube[4], -M_PI_2, 1, 0, 0);

	// bottom
	Rotate(cube[5], M_PI_2, 1, 0, 0);

	float angle = 0;
	const float ang_inc = 2.f * M_PI / 200.f;

	///////////////////////////////////////////////////////////////

	bigtime_t t0 = ::system_time();

	while (B_OK == (*frmwrk).frame_open(desc))
	{
		size_t pixel_chunk;
		size_t pixels_per_chunk;
		size_t row_alignment;

		::get_pixel_size_for(desc.pixelFormat, &pixel_chunk, &row_alignment, &pixels_per_chunk);

		int32 loop = desc.frameHeight;
		int32 bytes_per_row = (desc.frameWidth * pixel_chunk) / pixels_per_chunk;

		for (int8 *row = (int8 *) desc.pBits; loop--; row += desc.rowStride)
			::memset(row, 0x80, bytes_per_row);

		///////////////////////////////////////////////////////////////
		viewport.screen.left	= 0;
		viewport.screen.top		= 0;
		viewport.screen.right	= desc.frameWidth;
		viewport.screen.bottom	= desc.frameHeight;
	
		viewport.depth_range_n	= 0;
		viewport.depth_range_f	= 1;

		std::vector<BBitmap *> &mipmap = (*app).mTextureSet[(*app).mActiveTexSet];

		// fixup tex mapping {
		BBitmap *bitmapTex = *mipmap.begin();
		BRect boundsTex = (*bitmapTex).Bounds();
		const float repeat = 1.f;

		*(float *)&vt1[4] = (boundsTex.Width()	+ 1.f) * repeat - 1.f;
		*(float *)&vt3[4] = (boundsTex.Width()	+ 1.f) * repeat - 1.f;
		*(float *)&vt2[5] = (boundsTex.Height() + 1.f) * repeat - 1.f;
		*(float *)&vt3[5] = (boundsTex.Height() + 1.f) * repeat - 1.f;
		// } fixup tex mapping

		unsigned t;
		std::vector<BBitmap *>::iterator ti;

		for (ti = mipmap.begin(), t = 0; (ti != mipmap.end()) && (t != RENDER_STATE_MIPMAP_CAPACITY); ti++, t++)
		{
			BBitmap *bitmapTex = *ti;

			renderState.tex_lod[t].bits_const	= (const int8 *) (*bitmapTex).Bits();
			renderState.tex_lod[t].dimx			= (*bitmapTex).Bounds().IntegerWidth() + 1;
			renderState.tex_lod[t].dimy			= (*bitmapTex).Bounds().IntegerHeight() + 1;
			renderState.tex_lod[t].row_stride	= (*bitmapTex).BytesPerRow();
		}

		renderState.tex_min_lod = 0;
		renderState.tex_max_lod = t - 1;
		
		renderState.frame.bits			= (int8 *) desc.pBits;
		renderState.frame.row_stride	= desc.rowStride;
		renderState.frame.pix_format	= ::cfmapper.NativeToRender(desc.pixelFormat);

		renderState.culling = RENDER_STATE_CULL_CCW;
		renderState.sampling = (*app).mParams.rendSampling;

		Matrix4 t1;

		Rotate(t1, angle, 1, 1, -1);
		angle = ::fmodf(angle + ang_inc, 2.f * M_PI);

		for (unsigned i = 0; i < sizeof(cube) / sizeof(cube[0]); i++)
		{
			Matrix4 t0, t2;

			MulMat4(t1, cube[i], t0);

			static const float sqrt3 = ::sqrtf(3);

			Translate(t2, 0, 0, -(1.f + sqrt3));
			MulMat4(t2, t0);

			Frustum(t2, -1, 1, -1, 1, 1, 5);
			MulMat4(t2, t0);

			for (unsigned facet = 0; facet < sizeof(squareA) / sizeof(squareA[0]); facet++)
			{
				renderState.color_op = RENDER_STATE_OP_COPY;
				renderState.vert_format	= RENDER_STATE_VERT_UV;

				PlotTriangle6(t0,
								*(const Vector6 *)squareA[facet][0],
								*(const Vector6 *)squareA[facet][1],
								*(const Vector6 *)squareA[facet][2]);

				renderState.color_op = RENDER_STATE_OP_MODULATE;
				renderState.vert_format	= RENDER_STATE_VERT_ARGB;

				PlotTriangle8(t0,
								*(const Vector8 *)squareB[facet][0],
								*(const Vector8 *)squareB[facet][1],
								*(const Vector8 *)squareB[facet][2]);
			}
		}

		// frame completed, commit it ///////////////////////////////////
		(*frmwrk).frame_close();
		(*app).mFrameCount++;
	}

	(*app).mElapsedTime = ::system_time() - t0;

	return B_OK; 
}

// main /////////////////////////////////////////////////////////////////

int main()
{
	status_t error = B_OK;

	new(std::nothrow) App(error);

	if (NULL != ::be_app)
	{
		if (B_OK == error)	// situation normal: pass control to the app
			::be_app->Run();
		else 				// something went wrong at app construction
			std::cerr << "failure at constructing the application object." << std::endl;

		delete ::be_app;	// destroy app object and deallocate its space
	}
	else 					// failed to allocate space for the app object
	{
		std::cerr << "failed to allocate memory for the application object." << std::endl;
		error = B_ERROR;
	}

	return (int) error;
}
